''' Simplified .aup3 file interface '''

from .xml import encode, decode, get_root, make_root, Element
from .project import ElementSerializer, Project, SampleFormat
import sqlite3

__all__ = ['AUP3']

EXPECTED_DOCTYPE = '''
<?xml version="1.0" standalone="no" ?>
<!DOCTYPE project PUBLIC "-//audacityproject-1.3.0//DTD//EN" "http://audacity.sourceforge.net/xml/audacityproject-1.3.0.dtd" >
'''.strip()

FORMAT_DATA = {
	SampleFormat.INT16: ('int16', 0x8000),
	SampleFormat.INT24: ('int32', 0x800000),
	SampleFormat.FLOAT: ('float32', 1),
}

class AUP3:
	def __init__(self, path: str, ignore_autosave=False):
		'''
		opens an .aup3 file and (by default) performs some initial health checks.
		if you're going to modify the file (e.g. by writing to `[raw_]project`)
		it's highly recommended that the file isn't open in audacity at the same time.
		'''
		self.conn = sqlite3.Connection(path)
		assert self.conn.execute('SELECT count(*) FROM project').fetchone()[0] == 1
		if not ignore_autosave:
			assert (n := self.conn.execute('SELECT count(*) FROM autosave').fetchone()[0]) == 0, \
				f'{n} autosave records found'

	@property
	def raw_project(self) -> Element:
		'''
		project structure, as parsed XML tree. unlike `project`, this will always work
		unless the XML binary serialization format has changed, and re-serialization is
		almost byte-for-byte identical (save for [the order in] which strings are interned).

		note: for forensic use cases where absolute byte preservation is required,
		you might want to use the `xml` module's `TokenSerializer` class directly.
		'''
		names, doc = self.conn.execute('SELECT dict, doc FROM project').fetchone()
		doctype, root = get_root(decode(names, doc))
		assert doctype.strip() == EXPECTED_DOCTYPE,	f'unexpected doctype: {doctype!r}'
		return root

	@raw_project.setter
	def raw_project(self, value: Element):
		names, doc = encode(make_root(value, EXPECTED_DOCTYPE))
		res = self.conn.execute('UPDATE project SET dict = ?, doc = ?', [names, doc])
		assert res.rowcount == 1
		self.conn.commit()

	@property
	def project(self) -> Project:
		'''
		like `raw_project` but the XML data is validated and unmarshalled into Python
		dataclass objects. because Audacity doesn't have a well documented schema (its .dtd
		is hilariously out of date) this module makes no effort at backcompat and will
		raise an exception upon encountering *any* unknown attribute; in that case you
		may need to modify the dataclass types in `audacity_project` or use `raw_project`.

		this API is also a bit more lossy than `raw_project`: re-serialization does not
		preserve the order of attributes as well as child nodes belonging to different attributes,
		and the amount of significant digits is replaced with 7 (floats) and 19 (doubles)
		like in audacity-project-tools. also, the serializer is more lenient than the
		de-serializer as it assumes you're statically checking your Python code.
		'''
		return ElementSerializer.get(Project).deserialize(self.raw_project) # type: ignore

	@project.setter
	def project(self, value: Project):
		self.raw_project = ElementSerializer.get(Project).serialize(value)

	def get_block(self, id: int, normalized=True):
		'''
		returns sampleblock as a numpy array, optionally normalized to -1..+1 floats.
		WARNING: array must be manually reshaped according to channels
		'''
		import numpy
		format, samples = self.conn.execute('SELECT sampleformat, samples FROM sampleblocks WHERE blockid = ?', [id]).fetchone()
		dtype, factor = FORMAT_DATA[format]
		samples = numpy.frombuffer(samples, dtype=dtype)
		if normalized:
			samples = samples.astype('float32')
			samples /= factor
		return samples

	# TODO: add_block, set_block (maybe keep sampleformat by default)

	def close(self):
		''' explicitly close the .aup3 file (done automatically when self is deleted) '''
		self.conn.close()



''' Audacity binary XML serializer / deserializer '''

import struct
from typing import Union, Literal, Iterable, Iterator, get_args, NamedTuple
from dataclasses import dataclass, field
from itertools import tee
from io import BytesIO

lookahead = lambda it, default: next(tee(it, 1)[0], default)

__all__ = [
	'Attribute', 'Token', 'TokenSerializer',
	'Element', 'TextNode', 'Node', 'NodeDeserializer', 'NodeSerializer',
	'decode', 'encode', 'get_root',
]

# low-level token stream (lossless)

NameID = int # u16

Attribute = Union[
	tuple[Literal['String'], str],
	tuple[Literal['Int'], int], # i32
	tuple[Literal['Bool'], bool],
	tuple[Literal['Long'], int], # i32
	tuple[Literal['LongLong'], int], # i64
	tuple[Literal['SizeT'], int], # u32
	# for float/double there's an extra 'digits' i32
	tuple[Literal['Float'], float, int],
	tuple[Literal['Double'], float, int],
]

Token = Union[
	tuple[Literal['CharSize'], int], # u8
	tuple[Literal['StartTag'], NameID],
	tuple[Literal['EndTag'], NameID],
	tuple[Literal['Attr'], NameID, Attribute],
	tuple[Literal['Data'], str],
	tuple[Literal['Raw'], str],
	tuple[Literal['Push']],
	tuple[Literal['Pop']],
	tuple[Literal['Name'], NameID, str],
]

def read_fixed(f, n):
	data = f.read(n)
	assert len(data) == n, f'expected {n} bytes, got {len(data)}'
	return data

read_primitive = lambda f, fmt: \
	struct.unpack('<' + fmt, f.read(struct.Struct(fmt).size))[0]
write_primitive = lambda value, fmt: \
	struct.pack('<' + fmt, value)

read_bytes = lambda f, wide=True: \
	read_fixed(f, read_primitive(f, 'I' if wide else 'H'))
write_bytes = lambda value, wide=True: \
	[write_primitive(len(value), 'I' if wide else 'H'), value]

encodings = {1: 'utf-8', 2: 'utf-16le', 4: 'utf-32le'}
read_string = lambda f, nchars, wide=True: \
	read_bytes(f, wide=wide).decode(encodings[nchars])
write_string = lambda value, nchars, wide=True: \
	write_bytes(value.encode(encodings[nchars]), wide=wide)

read_bool = lambda f: {0: False, 1: True}[f.read(1)[0]]
write_bool = lambda value: [ bytes([ int(value) ]) ]

attr_names = [ get_args(get_args(t)[0])[0] for t in get_args(Attribute) ]
opcode_names = [ get_args(get_args(t)[0])[0] for t in get_args(Token) ]
opcode_names = sum(( attr_names if name == 'Attr' else [name] for name in opcode_names ), [])
name_opcodes = { name: ftype for ftype, name in enumerate(opcode_names) }

serializers = [
	'B', None, None,
	'string', 'i', (read_bool, write_bool), 'i', 'q', 'I', 'f', 'd',
	'string', 'string', None, None, 'string',
]

@dataclass
class TokenSerializer:
	nchars = -1

	def handle_token(self, token: Token):
		if token[0] == 'CharSize':
			assert token[1] in {1, 2, 4}, f'invalid char size {token[1]}'
			self.nchars = token[1]

	def read_one(self, f) -> Token:
		ftype = next(iter(f.read(1)))
		token = opcode_names[ftype],

		if 1 <= ftype < 11 or ftype == 15:
			token += read_primitive(f, 'H'),

		if handler := serializers[ftype]:
			if handler == 'string':
				token += read_string(f, self.nchars, ftype != 15),
			elif type(handler) is str:
				token += read_primitive(f, handler),
			else:
				token += handler[0](f),

		if 9 <= ftype < 11:
			token += read_primitive(f, 'i'),

		if 3 <= ftype < 11:
			token = 'Attr', token[1], (token[0], *token[2:]) # type: ignore
		self.handle_token(token)
		return token

	def write_one(self, token: Token) -> Iterable[bytes]:
		self.handle_token(token)
		if token[0] == 'Attr':
			token = token[2][0], token[1], *token[2][1:] # type: ignore

		name, *token = token # type: ignore
		ftype = name_opcodes[name]
		yield write_primitive(ftype, 'B')

		if 1 <= ftype < 11 or ftype == 15:
			fid, *token = token # type: ignore
			yield write_primitive(fid, 'H')

		if handler := serializers[ftype]:
			value, *token = token # type: ignore
			if handler == 'string':
				yield from write_string(value, self.nchars, ftype != 15)
			elif type(handler) is str:
				yield write_primitive(value, handler)
			else:
				yield from handler[1](value)

		if 9 <= ftype < 11:
			unk, *token = token # type: ignore
			yield write_primitive(unk, 'i')

		assert not token

	def read(self, f) -> Iterable[Token]:
		while True:
			try:
				yield self.read_one(f)
			except StopIteration:
				break

	def write(self, tokens: Iterable[Token]) -> Iterable[bytes]:
		for token in tokens:
			yield from self.write_one(token)

# high-level (does interning, tree structure)

class Element(NamedTuple):
	tag: str
	attrs: dict[str, Attribute]
	children: list['Node']

	def __repr__(self, prefix='') -> str:
		children = '\n' + ''.join(
			(c.__repr__(prefix + '    ') if type(c) is Element else repr(c)) + ',\n'
			for c in self.children) + prefix
		return prefix + f'Element({self.tag!r}, {self.attrs!r}, [{children if self.children else ""}])'

class TextNode(NamedTuple):
	is_raw: bool
	text: str

Node = Element | TextNode

@dataclass
class NodeDeserializer:
	names: dict[int, str] = field(default_factory=dict)

	def read_element(self, tag: str, rest: Iterator[Token]) -> Element:
		attrs = {}
		while (token := lookahead(rest, None)) and token[0] == 'Attr':
			next(rest)
			name = self.names[token[1]]
			assert name not in attrs
			attrs[name] = token[2]
		return Element(tag, attrs, list(self.read_nodes(rest, tag)))

	def read_one(self, token: Token, rest: Iterator[Token]) -> Node | None:
		if token[0] == 'Name':
			self.names[token[1]] = token[2]
		elif token[0] == 'Raw':
			return TextNode(True, token[1])
		elif token[0] == 'Data':
			return TextNode(False, token[1])
		elif token[0] == 'StartTag':
			return self.read_element(self.names[token[1]], rest)
		elif token[0] != 'CharSize':
			raise AssertionError(f'unexpected {token[0]} token')

	def read_nodes(self, tokens: Iterator[Token], tag=None) -> Iterable[Node]:
		for token in tokens:
			if token[0] == 'EndTag':
				assert self.names[token[1]] == tag, 'invalid end tag'
				return
			if node := self.read_one(token, tokens):
				yield node
		assert tag == None, 'missing end tag'

	def read(self, tokens: Iterable[Token]) -> Iterable[Node]:
		return self.read_nodes(tee(tokens, 1)[0])

@dataclass
class NodeSerializer:
	names: dict[str, int] = field(default_factory=dict)

	def write_name(self, name: str) -> Iterable[Token]:
		if self.names.setdefault(name, newid := len(self.names)) == newid:
			yield 'Name', newid, name

	def write_names(self, node: Node) -> Iterable[Token]:
		if isinstance(node, TextNode):
			return
		yield from self.write_name(node.tag)
		for key in node.attrs:
			yield from self.write_name(key)
		for child in node.children:
			yield from self.write_names(child)

	def write_doc(self, node: Node) -> Iterable[Token]:
		if isinstance(node, TextNode):
			yield ('Raw' if node.is_raw else 'Data'), node.text # type: ignore
			return
		yield 'StartTag', self.names[node.tag]
		for key, value in node.attrs.items():
			yield 'Attr', self.names[key], value
		for child in node.children:
			yield from self.write_doc(child)
		yield 'EndTag', self.names[node.tag]

# convenience API that combines the two layers

def decode(names: bytes, doc: bytes) -> list[Node]:
	td = TokenSerializer()
	nd = NodeDeserializer()
	nodes = list(nd.read(td.read(BytesIO(names))))
	assert not nodes, 'dictionary portion must not have content'
	return list(nd.read(td.read(BytesIO(doc))))

def concat(chunks: Iterable[bytes]) -> bytes:
	st = BytesIO()
	for chunk in chunks:
		st.write(chunk)
	return bytes(st.getbuffer())

def encode(nodes: list[Node]) -> tuple[bytes, bytes]:
	ts = TokenSerializer()
	ns = NodeSerializer()
	names = list(ts.write_one( ('CharSize', 4) ))
	for node in nodes:
		names += ts.write(ns.write_names(node))
	doc = []
	for node in nodes:
		doc += ts.write(ns.write_doc(node))
	return concat(names), concat(doc)

# add / remove doctype and xmlns

XMLNS = 'http://audacity.sourceforge.net/xml/'

def get_root(nodes: list[Node], xmlns=XMLNS) -> tuple[str, Element]:
	doctype = ''
	while nodes and isinstance(nodes[0], TextNode) and nodes[0].is_raw:
		doctype += nodes.pop(0).text  # type: ignore

	assert len(nodes) == 1
	root, = nodes
	assert isinstance(root, Element)
	root_xmlns = root.attrs.pop('xmlns', None)
	assert root_xmlns == ('String', xmlns), f'unexpected namespace {root_xmlns}'
	return doctype, root

def make_root(root: Element, doctype: str, xmlns=XMLNS) -> list[Node]:
	value = Element(root.tag, { **root.attrs, 'xmlns': ('String', xmlns) }, root.children)
	return [ TextNode(True, doctype), value ]



''' Audacity project schema unmarshaller '''

from . import xml
from dataclasses import dataclass, is_dataclass
from typing import Annotated, Optional, Literal, Union, get_origin, get_args
from types import UnionType, NoneType
from enum import IntEnum, unique
from collections import defaultdict

@dataclass(frozen=True)
class AudacityAttribute:
	kind: str

Long = Annotated[int, AudacityAttribute('Long')] # i32
LongLong = Annotated[int, AudacityAttribute('LongLong')] # i64
SizeT = Annotated[int, AudacityAttribute('SizeT')] # u32
Double = Annotated[float, AudacityAttribute('Double')]

_attribute_names = { 'String': str, 'Int': int, 'Bool': bool, 'Long': Long, 'LongLong': LongLong, 'SizeT': SizeT, 'Float': float, 'Double': Double }

class BaseElement: pass

ChildAttrModifier = Literal['one', 'many', 'optional']

class ElementSerializer:
	@classmethod
	def is_child_attr(cls, at: type) -> Optional[tuple[ChildAttrModifier, list[type]]]:
		origin, args = get_origin(at), get_args(at)
		if ct := cls.is_direct_child_attr(at):
			return 'one', ct
		if origin is list and len(args) == 1 and (ct := cls.is_direct_child_attr(args[0])):
			return 'many', ct
		if origin is Union and len(args) == 2 and args[1] is NoneType and (ct := cls.is_direct_child_attr(args[0])):
			return 'optional', ct

	@classmethod
	def is_direct_child_attr(cls, at: type) -> Optional[list[type]]:
		origin, args = get_origin(at), get_args(at)
		if type(at) is type and issubclass(at, BaseElement):
			return [at]
		if (origin is Union or origin is UnionType) and all(type(t) is type and issubclass(t, BaseElement) for t in args):
			return list(args)

	def __init__(self, element_cls: type):
		assert is_dataclass(element_cls) and issubclass(element_cls, BaseElement)
		self.cls = element_cls
		self.tag = element_cls.__name__.lower()
		self.children: dict[str, tuple[ChildAttrModifier, dict[type, ElementSerializer]]] = {}
		self.attrs = {}

		for key, atype in element_cls.__annotations__.items():
			if ct := self.is_child_attr(atype):
				tags = { t: self.get(t) for t in ct[1] }
				self.children[key] = ct[0], tags
				continue
			origin, args = get_origin(atype), get_args(atype)
			optional = False
			if origin is Union and len(args) == 2 and args[1] is NoneType:
				atype = args[0]
				optional = True
			kind = next((k for k, t in _attribute_names.items() if atype is t), None)
			assert kind, f'failed to parse attribute {key}: {atype}'
			self.attrs[key] = optional, kind

		children_by_tag = [ (t.tag, (t, k)) for k, (_, tags) in self.children.items() for t in tags.values() ]
		self.children_by_tag = dict(children_by_tag)
		assert len(children_by_tag) == len(children_by_tag)

	def deserialize(self, el: xml.Element, strict=True):
		assert el.tag == self.tag
		args = {}
		attrs = el.attrs.copy()
		for key, (optional, kind) in self.attrs.items():
			if value := attrs.pop(key, None):
				assert value[0] == kind, f'attribute {key} expected to be {kind}, got {value[0]}'
				args[key] = value[1]
			else:
				assert optional, f'required attribute {key} not present'
		assert (not strict) or (not attrs), f'unexpected attributes: {attrs}'

		children = defaultdict(list)
		for node in el.children:
			assert not isinstance(node, xml.TextNode)
			if entry := self.children_by_tag.get(node.tag):
				children[entry[1]].append(entry[0].deserialize(node, strict=strict))
			else:
				assert not strict, f'unexpected {node.tag!r} element'
		for key, (modifier, _) in self.children.items():
			nodes = children[key]
			if modifier == 'one':
				assert len(nodes) == 1, f'expected one of {key}, got {len(nodes)}'
				nodes = nodes[0]
			elif modifier == 'optional':
				assert len(nodes) <= 1, f'expected at most one of {key}, got {len(nodes)}'
				nodes = nodes[0] if nodes else None
			else:
				assert modifier == 'many'
			args[key] = nodes

		return self.cls(**args)

	def serialize(self, el) -> xml.Element:
		attrs = {}
		for key, (_, kind) in self.attrs.items():
			if (value := getattr(el, key)) != None:
				attrs[key] = (kind, value) + {'Float': (7,), 'Double': (19,)}.get(kind, ())
		children = []
		for key, (kind, serializers) in self.children.items():
			value = getattr(el, key)
			if kind != 'many':
				value = [value] if value != None else []
			children += ( serializers[type(el)].serialize(el) for el in value )
		return xml.Element(self.tag, attrs, children)

	__serializers = {}

	@staticmethod
	def get(element_cls: type):
		if (x := ElementSerializer.__serializers.get(element_cls)) == None:
			x = ElementSerializer.__serializers[element_cls] = ElementSerializer(element_cls)
		return x

# schema types:

ProjectTime = Double # seconds, 10 digits

@unique
class SampleFormat(IntEnum):
	INT16 = 0x20001
	INT24 = 0x40001
	FLOAT = 0x4000F

TagName = Literal[
	'COMMENTS',
	'GENRE',
	'YEAR',
	'TRACKNUMBER',
	'ALBUM',
	'TITLE',
	'ARTIST',
]

@dataclass
class Tag(BaseElement):
	name: str
	value: str

@dataclass
class Tags(BaseElement):
	tags: list[Tag]

@dataclass
class Effects(BaseElement):
	active: bool
	# TODO

@dataclass
class Label(BaseElement):
	t: ProjectTime
	t1: ProjectTime
	title: str
	selLow: ProjectTime
	selHigh: ProjectTime

@dataclass
class LabelTrack(BaseElement):
	name: str
	isSelected: bool
	height: int # px
	minimized: bool

	numlabels: int
	labels: list[Label]

@dataclass
class ControlPoint(BaseElement):
	t: ProjectTime
	val: Double

@dataclass
class Envelope(BaseElement):
	numpoints: SizeT
	points: list[ControlPoint]

@dataclass
class TimeTrack(BaseElement):
	name: str
	isSelected: bool
	height: int # px
	minimized: bool

	rangelower: Double # 0.2, 12
	rangeupper: Double # 2.0, 12
	displaylog: bool
	interpolatelog: bool

	envelope: Envelope

@dataclass
class WaveBlock(BaseElement):
	start: LongLong # sample number
	blockid: LongLong

@dataclass
class Sequence(BaseElement):
	maxsamples: SizeT
	numsamples: LongLong
	sampleformat: SizeT # SampleFormat
	effectivesampleformat: SizeT # SampleFormat
	blocks: list[WaveBlock]

@dataclass
class WaveClip(BaseElement):
	offset: ProjectTime
	trimLeft: ProjectTime
	trimRight: ProjectTime
	centShift: int
	pitchAndSpeedPreset: Long
	rawAudioTempo: Double # bpm
	clipStretchRatio: Double
	name: str
	colorindex: int

	sequence: Sequence
	envelope: Envelope

@dataclass
class WaveTrack(BaseElement):
	name: str
	isSelected: bool
	height: int # px
	minimized: bool

	colorindex: int # 0
	channel: int # 0
	linked: int # 0
	mute: bool
	solo: bool
	rate: Double # 44100.0, -1
	gain: Double # 1.0, -1
	pan: Double # 0.0, -1
	sampleformat: Long # SampleFormat
	# offset: ProjectTime ?

	clips: list[WaveClip]
	effects: Optional[Effects] = None

	# for stereo tracks:
	height1: Optional[int] = None
	minimized1: Optional[bool] = None

@dataclass
class Project(BaseElement):
	version: str # '1.3.0'
	audacityversion: str # '3.7.3'

	rate: Double # Hz, -1
	selectionformat: str # 'hh:mm:ss + milliseconds'
	frequencyformat: str # 'Hz'
	bandwidthformat: str # 'octaves'
	time_signature_tempo: Double # bpm, -1
	time_signature_upper: int # 4
	time_signature_lower: int # 4
	snapto: str # 'off'
	sel0: ProjectTime
	sel1: ProjectTime
	selLow: ProjectTime
	selHigh: ProjectTime
	vpos: int # 0
	h: Double # 0.0, 10
	zoom: ProjectTime
	preferred_export_rate: Double # Hz or 0, -1

	effects: Effects
	tags: Tags
	tracks: list[WaveTrack | LabelTrack | TimeTrack]

ElementSerializer.get(Project)
